import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers, Model
from tensorflow.keras.utils import register_keras_serializable, serialize_keras_object, deserialize_keras_object
import numpy as np

def create_padding_mask(seq):
    mask = tf.cast(tf.math.equal(seq, 0), tf.float32)
    return mask[:, tf.newaxis, tf.newaxis, :]

# Класс внимания мульти-голов
@register_keras_serializable(package='Custom', name='CustomMultiHeadAttention')
class CustomMultiHeadAttention(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads):
        super(CustomMultiHeadAttention, self).__init__()
        self.num_heads = num_heads
        self.d_model = d_model
        assert d_model % self.num_heads == 0
        self.depth = d_model // self.num_heads # глубина - размерность выходного вектора после умножения матрицы эмбеддингов на Q,K,V
        self.wq = tf.keras.layers.Dense(d_model)
        self.wk = tf.keras.layers.Dense(d_model)
        self.wv = tf.keras.layers.Dense(d_model)
        self.dense = tf.keras.layers.Dense(d_model)

    def split_heads(self, x, batch_size):
        # -1 означает, что длинна последовательности остается прежней
        x = tf.reshape(x, (batch_size, -1, self.num_heads, self.depth))  # кол-во списков batch_size(не изменяется), кол-во последовательностей не изменено(токенов), кол-во разбиений эмбеддингов слов изменено на num_heads(кол-во списков в колове), глубина головы(кол-во (весов+позиция) в одном списке) 
        return tf.transpose(x, perm=[0, 2, 1, 3])

    def call(self, v, k, q, mask):
        batch_size = tf.shape(q)[0]
        q = self.wq(q)  # (кол-во обучающих предложений, длинна последовательности, размерность вектора для токена в эмбеддинге(d_model)) матрица запросов
        k = self.wk(k)  # (кол-во обучающих предложений, длинна последовательности, размерность вектора для токена в эмбеддинге(d_model)) матрица, на что мы хотим обратить внимание
        v = self.wv(v)  # (кол-во обучающих предложений, длинна последовательности, размерность вектора для токена в эмбеддинге(d_model))
        q = self.split_heads(q, batch_size)  # (batch_size, num_heads, seq_len_q, depth)
        k = self.split_heads(k, batch_size)  # (batch_size, num_heads, seq_len_k, depth)
        v = self.split_heads(v, batch_size)  # (batch_size, num_heads, seq_len_v, depth)

        # внимание к масштабированному скалярному произведению(scaled dot product attention)
        matmul_qk = tf.matmul(q, k, transpose_b=True) # transpose(0, 1, 3, 2) (batch_size, num_heads, seq_len_q, seq_len_k)
        dk = tf.cast(tf.shape(k)[-1], tf.float32) # получение размерности 
        scaled_attention_logits = matmul_qk / tf.math.sqrt(dk)
        if mask is not None:
            scaled_attention_logits += (mask * -1e9)
        attention_weights = tf.nn.softmax(scaled_attention_logits, axis=-1)
        scaled_attention = tf.matmul(attention_weights, v)
        scaled_attention = tf.transpose(scaled_attention, perm=[0, 2, 1, 3])
        concat_attention = tf.reshape(scaled_attention, (batch_size, -1, self.d_model))
        output = self.dense(concat_attention)
        return output, attention_weights
        

    def get_config(self):
        config = super(CustomMultiHeadAttention, self).get_config()
        config.update({
            'num_heads': self.num_heads,
            'd_model': self.d_model,
        })
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)

# слой энкодера
@register_keras_serializable(package='Custom', name='EncoderLayer')
class EncoderLayer(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads, dff, rate, SEED, **kwargs):
        super(EncoderLayer, self).__init__(**kwargs)

        self.mha = CustomMultiHeadAttention(d_model, num_heads)
        self.ffn = tf.keras.Sequential([
            tf.keras.layers.Dense(dff, activation='relu', kernel_initializer='he_normal'),
            tf.keras.layers.Dense(d_model,kernel_regularizer=tf.keras.regularizers.l2(1e-2))
        ])

        self.layernorm1 = tf.keras.layers.LayerNormalization(epsilon=1e-2)
        self.layernorm2 = tf.keras.layers.LayerNormalization(epsilon=1e-2)
        self.dropout1 = tf.keras.layers.Dropout(rate, seed=SEED)
        self.dropout2 = tf.keras.layers.Dropout(rate, seed=SEED)
        

    def call(self, x, training, mask):
        attn_output, attn_weights = self.mha(x, x, x, mask) # подается и выходит attantion score
        attn_output = self.dropout1(attn_output, training=training) # слой dropout сразу после слоя Dence из multi-head attention
        out1 = self.layernorm1(x + attn_output) # здесь attantion score складываеться с начальной матрицей эмбеддингов, чтоб обеспечить более сильный сигнал связи между начальными значениями и выходными.
        ffn_output = self.ffn(out1)
        ffn_output = self.dropout2(ffn_output, training=training)
        out2 = self.layernorm2(out1 + ffn_output)
        return out2, attn_weights

    def get_config(self):
        config = super(EncoderLayer, self).get_config()
        config.update({
            "d_model": self.mha.d_model,
            "num_heads": self.mha.num_heads,
            "dff": self.ffn.layers[0].units,
            "rate": self.dropout1.rate,
            "SEED": self.dropout1.seed
        })
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)

# класс энкодера

@register_keras_serializable(package='Custom', name='Encoder')
class Encoder(tf.keras.layers.Layer):
    def __init__(self, num_layers, d_model, num_heads, dff, input_vocab_size, maximum_pos_encoding, rate, SEED, **kwargs):
        super(Encoder, self).__init__(**kwargs)
        self.num_layers = num_layers
        self.d_model = d_model
        self.embedding = tf.keras.layers.Embedding(input_vocab_size, d_model) # входной слой
        self.pos_encoding = self.positional_encoding(maximum_pos_encoding, d_model)
        self.enc_layers = [EncoderLayer(d_model, num_heads, dff, rate, SEED) for _ in range(num_layers)]
        self.dropout = tf.keras.layers.Dropout(rate, seed=SEED)
        

    def positional_encoding(self, position, d_model):
        def get_angles(pos, i, d_model):
            angle_rates = 1 / np.power(10000, (2 * (i // 2)) / np.float32(d_model))
            return pos * angle_rates

        angle_rads = get_angles(np.arange(position)[:, np.newaxis], np.arange(d_model)[np.newaxis, :], d_model)
        angle_rads[:, 0::2] = np.sin(angle_rads[:, 0::2]) # четные
        angle_rads[:, 1::2] = np.cos(angle_rads[:, 1::2]) # не четные
        pos_encoding = angle_rads[np.newaxis, ...]
        return tf.cast(pos_encoding, dtype=tf.float32)

    def call(self, x, training, mask):
        seq_len = tf.shape(x)[1] # получение длины последовательности (кол-ва токенов в одном предложении)
        x = self.embedding(x) # построение матрицы эмбеддингов
        x *= tf.math.sqrt(tf.cast(self.d_model, tf.float32))
        x += self.pos_encoding[:, :seq_len, :] # применение позиционного кодирования только к токенам
        x = self.dropout(x, training=training)
        
        attention_weights = {}
        for i in range(self.num_layers):
            x, attn_weights = self.enc_layers[i](x, training, mask) # слои энкодера по порядку передают значение X, обучая матрицы
            attention_weights[f'layer_{i + 1}_block_1'] = attn_weights
        
        return x, attention_weights

    def get_config(self):
        config = super(Encoder, self).get_config()
        config.update({
            'num_layers': self.num_layers,
            'd_model': self.d_model,
            'num_heads': self.enc_layers[0].mha.num_heads,
            'dff': self.enc_layers[0].ffn.layers[0].units,
            'input_vocab_size': self.embedding.input_dim,
            'maximum_pos_encoding': self.pos_encoding.shape[1],
            'rate': self.dropout.rate,
            'SEED': self.dropout.seed
        })
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)

#инициализация модели
@register_keras_serializable(package='Custom', name='ModifiedModel')
class ModifiedModel(Model):
    def __init__(self, encoder, num_classes, rate, SEED, **kwargs):
        super(ModifiedModel, self).__init__(**kwargs)
        self.encoder = encoder
        self.global_average_pooling = layers.GlobalAveragePooling1D() # Аггрегация представлений (Aggregation Layers)
        self.dropout = layers.Dropout(rate, seed=SEED)
        self.final_layer = layers.Dense(num_classes, activation='softmax')

        

    def call(self, inputs, training=None, mask=None):
        x, attention_weights = self.encoder(inputs[0], training, mask=inputs[1])
        x = self.global_average_pooling(x)
        x = self.dropout(x, training=training)
        x = self.final_layer(x)
        if not training:
            return x, attention_weights
        return x

    def predict_with_attention(self, inputs):
        predictions, attention_weights = self.call(inputs, training=False)
        return predictions, attention_weights

    def get_config(self):
        base_config = super(ModifiedModel, self).get_config()
        config = {
            "encoder": tf.keras.utils.serialize_keras_object(self.encoder),
            'num_classes': self.final_layer.units,
            'rate': self.dropout.rate,
            'SEED': self.dropout.seed
        }
        return {**base_config, **config}

    @classmethod
    def from_config(cls, config):
        encoder_config = config.pop('encoder')
        encoder = deserialize_keras_object(encoder_config)  
        return cls(encoder, **config)